# ORDERED TREES
#
# Represented using a parent-child relation using a dictionary: tree =
# { 'a' : ['b', 'c'], 'b' : ['d', 'e'], 'c' : [], ...} means that 'a'
# has two children 'b' and 'c', 'b' has two children 'd' and 'e', 'c'
# has no children, etc.
#
# The use of lists allows to capture the relative order between
# siblings. We talk about ordered trees.
#

if __name__ == '__main__':
    tree = {
        'a' : ['b', 'c'],
        'b' : ['d', 'e'],
        'c' : [],
        'd' : [],
        'e' : ['f', 'g', 'h'],
        'f' : [],
        'g' : ['i', 'j'],
        'h' : [],
        'i' : [],
        'j' : []
    }

def find_nodes(tree):
    return set(tree.keys())

if __name__ == '__main__':
    assert find_nodes(tree) == set(['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j'])

def find_inner_nodes(tree):
    ns = set()
    for parent in tree:
        for child in tree[parent]:
            ns.add(child)
    return ns

if __name__ == '__main__':
    assert find_inner_nodes(tree) == set(['b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j'])

def find_root(tree):
    roots = find_nodes(tree) - find_inner_nodes(tree)
    return list(roots)[0]

if __name__ == '__main__':
    find_root(tree) == 'a'

def explore_rec(tree, node, traversal):
    traversal.append(node)

    for child in tree[node]:
        explore_rec(tree, child, traversal)

    return traversal

def dfs_traversal(tree):
    root = find_root(tree)
    
    traversal = []
    explore_rec(tree, root, traversal)

    return traversal


if __name__ == '__main__':
    assert dfs_traversal(tree) == ['a', 'b', 'd', 'e', 'f', 'g', 'i', 'j', 'h', 'c']
