# UNORDERED TREES 
# Representing a tree with a child-parent relation using a dictionary
# tree = { 'A' : None, 'B' : 'A' ...} means that 'A' doesn't have a
# parent (is a root), 'B' has a parent 'A', etc.

# NB: this representation does not allow (innately) to store relative
# order among siblings. We talk about unordered tree.

if __name__ == '__main__':
    tree = {
        'A' : None, 
        'B' : 'A', 
        'C' : 'A', 
        'D' : 'B', 
        'E' : 'B', 
        'F' : 'E',
        'G' : 'E', 
        'H' : 'E', 
        'I' : 'G', 
        'J' : 'G',
    }

def find_root(tree):
    for node in tree:
        if tree[node] is None:
            return node

if __name__ == '__main__':
    assert find_root(tree) == 'A'


def find_children(tree, node):
    chs = set()
    for (c, p) in tree.items():
        if p == node:
            chs.add(c)
    return chs

if __name__ == '__main__':
    assert find_children(tree, 'A') == set(['B', 'C'])
    assert find_children(tree, 'E') == set(['F', 'G', 'H'])
    assert find_children(tree, 'C') == set([])
        

# unordered recursive exploration of a tree at a given node
def explore_rec(tree, node, traversal):
    traversal.append(node)
    for c in find_children(tree, node):
        explore_rec(tree, c, traversal)
    return traversal

def dfs_traversal(tree):
    root = find_root(tree)

    traversal = []
    explore_rec(tree, root, traversal)

    return traversal

if __name__ == '__main__':
    traversal = dfs_traversal(tree)
    # root is the first element in the traversal
    assert traversal.index(find_root(tree)) == 0


    # if node has children, then one of them has a number immediately
    # following the number of a parent
    for node in tree:
        if find_children(tree, node):
            assert traversal.index(node) + 1 == min([traversal.index(child)
                                                    for child in find_children(tree, node)])

